import os
import re
import subprocess
from pathlib import Path
from typing import Literal

import pytest
from apify_client import ApifyClientAsync
from cookiecutter.main import cookiecutter

from crawlee._cli import default_start_url, template_directory
from crawlee._utils.crypto import crypto_random_object_id
from tests.e2e.project_template.utils import patch_crawlee_version_in_project

# To run these tests locally, make sure you have apify-cli installed and available in the path.
# https://docs.apify.com/cli/docs/installation


@pytest.mark.parametrize(
    'crawler_type',
    [
        pytest.param('playwright-camoufox', marks=pytest.mark.playwright_camoufox),
        pytest.param('playwright-chrome', marks=pytest.mark.playwright_chrome),
        pytest.param('playwright-firefox', marks=pytest.mark.playwright_firefox),
        pytest.param('playwright-webkit', marks=pytest.mark.playwright_webkit),
        pytest.param('playwright', marks=pytest.mark.playwright),
        pytest.param('parsel', marks=pytest.mark.parsel),
        pytest.param('beautifulsoup', marks=pytest.mark.beautifulsoup),
    ],
)
@pytest.mark.parametrize(
    'http_client',
    [
        pytest.param('httpx', marks=pytest.mark.httpx),
        pytest.param('curl-impersonate', marks=pytest.mark.curl_impersonate),
        pytest.param('impit', marks=pytest.mark.impit),
    ],
)
@pytest.mark.parametrize(
    'package_manager',
    [
        pytest.param('pip', marks=pytest.mark.pip),
        pytest.param('uv', marks=pytest.mark.uv),
        pytest.param('poetry', marks=pytest.mark.poetry),
    ],
)
async def test_static_crawler_actor_at_apify(
    tmp_path: Path,
    crawlee_wheel_path: Path,
    package_manager: Literal['pip', 'uv', 'poetry'],
    crawler_type: str,
    http_client: str,
) -> None:
    # Generate new actor name
    actor_name = f'crawlee-python-template-e2e-test-{crypto_random_object_id(8).lower()}'

    # Create project from template
    cookiecutter(
        template=str(template_directory),
        no_input=True,
        extra_context={
            'project_name': actor_name,
            'package_manager': package_manager,
            'crawler_type': crawler_type,
            'http_client': http_client,
            'enable_apify_integration': True,
            'start_url': default_start_url,
            'install_project': False,
        },
        output_dir=str(tmp_path),
    )

    patch_crawlee_version_in_project(
        project_path=tmp_path / actor_name, wheel_path=crawlee_wheel_path, package_manager=package_manager
    )

    # Print apify version for debugging purposes in rare cases of CLI failures
    subprocess.run(['apify', '--version'], check=True)  # noqa: ASYNC221, S607

    # Build actor using sequence of cli commands as the user would
    subprocess.run(  # noqa: ASYNC221, S603
        ['apify', 'login', '-t', os.environ['APIFY_TEST_USER_API_TOKEN']],  # noqa: S607
        capture_output=True,
        check=True,
        cwd=tmp_path / actor_name,
    )
    subprocess.run(['apify', 'init', '-y', actor_name], capture_output=True, check=True, cwd=tmp_path / actor_name)  # noqa: ASYNC221, S603, S607

    build_process = subprocess.run(['apify', 'push'], capture_output=True, check=False, cwd=tmp_path / actor_name)  # noqa: ASYNC221, S607
    # Get actor ID from build log
    actor_id_regexp = re.compile(r'https:\/\/console\.apify\.com\/actors\/(.*)#\/builds\/\d*\.\d*\.\d*')

    if match := re.findall(actor_id_regexp, build_process.stderr.decode()):
        actor_id = match[0]
    else:
        raise AssertionError(f'Failed to find actor id in build log: {build_process.stderr.decode()}')

    client = ApifyClientAsync(token=os.getenv('APIFY_TEST_USER_API_TOKEN'))
    actor = client.actor(actor_id)

    # Run actor
    try:
        assert build_process.returncode == 0
        started_run_data = await actor.start(memory_mbytes=8192)
        actor_run = client.run(started_run_data['id'])

        finished_run_data = await actor_run.wait_for_finish()
        actor_run_log = await actor_run.log().get()
    finally:
        # Delete the actor once it is no longer needed.
        await actor.delete()

    # Asserts
    additional_run_info = f'Full actor run log: {actor_run_log}'
    assert actor_run_log
    assert finished_run_data
    assert finished_run_data['status'] == 'SUCCEEDED', additional_run_info
    assert (
        'Crawler.stop() was called with following reason: The crawler has reached its limit of 10 requests per crawl.'
    ) in actor_run_log, additional_run_info
    assert int(re.findall(r'requests_finished\s*│\s*(\d*)', actor_run_log)[-1]) >= 10, additional_run_info



import asyncio

from crawlee.crawlers import BeautifulSoupCrawler, BeautifulSoupCrawlingContext


async def main() -> None:
    crawler = BeautifulSoupCrawler()

    @crawler.router.default_handler
    async def request_handler(context: BeautifulSoupCrawlingContext) -> None:
        url = context.request.url
        title = context.soup.title.string if context.soup.title else ''
        context.log.info(f'The title of {url} is: {title}.')

    await crawler.run(['https://crawlee.dev/'])


if __name__ == '__main__':
    asyncio.run(main())



import asyncio

from crawlee import Glob
from crawlee.crawlers import BeautifulSoupCrawler, BeautifulSoupCrawlingContext


async def main() -> None:
    crawler = BeautifulSoupCrawler(max_requests_per_crawl=10)

    @crawler.router.default_handler
    async def request_handler(context: BeautifulSoupCrawlingContext) -> None:
        context.log.info(f'Processing {context.request.url}.')

        # Enqueue links that match the 'include' glob pattern and
        # do not match the 'exclude' glob pattern.
        # highlight-next-line
        await context.enqueue_links(
            # highlight-next-line
            include=[Glob('https://someplace.com/**/cats')],
            # highlight-next-line
            exclude=[Glob('https://**/archive/**')],
            # highlight-next-line
        )

    await crawler.run(['https://crawlee.dev/'])


if __name__ == '__main__':
    asyncio.run(main())



import asyncio

from crawlee.crawlers import BeautifulSoupCrawler, BeautifulSoupCrawlingContext
from crawlee.storages import Dataset


async def main() -> None:
    # Open the dataset, if it does not exist, it will be created.
    # Leave name empty to use the default dataset.
    dataset = await Dataset.open(name='my-dataset')

    # Create a new crawler (it can be any subclass of BasicCrawler).
    crawler = BeautifulSoupCrawler()

    # Define the default request handler, which will be called for every request.
    @crawler.router.default_handler
    async def request_handler(context: BeautifulSoupCrawlingContext) -> None:
        context.log.info(f'Processing {context.request.url} ...')

        # Extract data from the page.
        data = {
            'url': context.request.url,
            'title': context.soup.title.string if context.soup.title else None,
        }

        # Push the extracted data to the dataset.
        await dataset.push_data(data)

    # Run the crawler with the initial URLs.
    await crawler.run(['https://crawlee.dev'])

    # Export the dataset to the key-value store.
    await dataset.export_to(key='dataset', content_type='csv')


if __name__ == '__main__':
    asyncio.run(main())



import traceback

import pytest

from crawlee.statistics._error_tracker import ErrorTracker


@pytest.mark.parametrize(
    ('error_tracker', 'expected_unique_errors'),
    [
        (ErrorTracker(), 5),
        (ErrorTracker(show_file_and_line_number=False), 4),
        (ErrorTracker(show_error_name=False), 4),
        (ErrorTracker(show_error_message=False), 3),
        (ErrorTracker(show_error_name=False, show_file_and_line_number=False), 3),
        (ErrorTracker(show_file_and_line_number=False, show_error_message=False), 2),
        (ErrorTracker(show_error_name=False, show_file_and_line_number=False, show_error_message=False), 1),
    ],
)
async def test_error_tracker_counts(error_tracker: ErrorTracker, expected_unique_errors: int) -> None:
    """Use different settings of `error_tracker` and test unique errors count."""

    for error in [
        Exception('Some value error abc'),
        ValueError('Some value error abc'),  # Different type, different error
        ValueError('Some value error cde'),  # Same type and similar message to previous, considered the same.
        ValueError(
            'Another value error efg'
        ),  # Same type, but too different message to previous, considered different.
        ValueError(),  # Same type but don't have message, considered different.
    ]:
        try:
            raise error  # Errors raised on same line
        except Exception as e:  # noqa:PERF203
            await error_tracker.add(e)

    try:
        raise ValueError('Some value error abc')  # Same as one previous error, but different line.
    except Exception as e:
        await error_tracker.add(e)

    assert error_tracker.total == 6
    assert error_tracker.unique_error_count == expected_unique_errors


@pytest.mark.parametrize(
    ('message_1', 'message_2', 'expected_generic_message'),
    [
        ('Some error number 123', 'Some error number 456', 'Some error number ***'),
        ('Some error number 123 456', 'Some error number 123 456 789', 'Some error number 123 456 ***'),
        ('Some error number 0 0 0', 'Some error number 1 0 1', 'Some error number *** 0 ***'),
    ],
)
async def test_error_tracker_similar_messages_full_stack(
    message_1: str, message_2: str, expected_generic_message: str
) -> None:
    """Test that similar messages collapse into same group with generic name that contains wildcard symbols."""
    error_tracker = ErrorTracker()
    for error in [
        KeyError(message_1),
        KeyError(message_1),
        KeyError(message_1),
        ValueError(message_1),
        ValueError(message_2),
        RuntimeError(message_2),
    ]:
        try:
            raise error  # Errors raised on the same line
        except Exception as e:  # noqa:PERF203
            await error_tracker.add(e)
            line = traceback.extract_tb(e.__traceback__)[0].lineno

    file_name = __file__.split('/')[-1]
    errors = error_tracker.get_most_common_errors()
    assert errors[0][0] == f'{file_name}:{line}:KeyError:{message_1}'
    assert errors[0][1] == 3
    assert errors[1][0] == f'{file_name}:{line}:ValueError:{expected_generic_message}'
    assert errors[1][1] == 2
    assert errors[2][0] == f'{file_name}:{line}:RuntimeError:{message_2}'
    assert errors[2][1] == 1


@pytest.mark.parametrize(
    ('show_full_message', 'expected_message'),
    [
        (True, 'Error line 1\n Error line 2'),
        (False, 'Error line 1'),
    ],
)
async def test_show_full_message(*, show_full_message: bool, expected_message: str) -> None:
    """Test error message settings with both options of `show_full_message`."""
    error_tracker = ErrorTracker(
        show_error_name=False, show_file_and_line_number=False, show_full_message=show_full_message
    )

    try:
        raise RuntimeError('Error line 1\n Error line 2')  # Errors raised on the same line
    except Exception as e:
        await error_tracker.add(e)

    assert error_tracker.get_most_common_errors()[0][0] == expected_message


async def test_error_tracker_with_errors_chain() -> None:
    """Test error tracker with errors chain."""
    error_tracker = ErrorTracker(show_error_name=False, show_file_and_line_number=False, show_full_message=True)

    try:
        raise ZeroDivisionError('Zero division error')  # Errors raised on the same line
    except Exception as e:
        try:
            raise ValueError from e
        except Exception as e:
            await error_tracker.add(e)

    assert error_tracker.get_most_common_errors()[0][0] == 'Zero division error'



import asyncio

from crawlee.crawlers import (
    PlaywrightCrawler,
    PlaywrightCrawlingContext,
    PlaywrightPreNavCrawlingContext,
)


async def main() -> None:
    crawler = PlaywrightCrawler(
        # Limit the crawl to max requests. Remove or increase it for crawling all links.
        max_requests_per_crawl=10,
        # Headless mode, set to False to see the browser in action.
        headless=False,
        # Browser types supported by Playwright.
        browser_type='chromium',
    )

    # Define the default request handler, which will be called for every request.
    # The handler receives a context parameter, providing various properties and
    # helper methods. Here are a few key ones we use for demonstration:
    # - request: an instance of the Request class containing details such as the URL
    #   being crawled and the HTTP method used.
    # - page: Playwright's Page object, which allows interaction with the web page
    #   (see https://playwright.dev/python/docs/api/class-page for more details).
    @crawler.router.default_handler
    async def request_handler(context: PlaywrightCrawlingContext) -> None:
        context.log.info(f'Processing {context.request.url} ...')

        # Extract data from the page using Playwright's API.
        posts = await context.page.query_selector_all('.athing')
        data = []

        for post in posts:
            # Get the HTML elements for the title and rank within each post.
            title_element = await post.query_selector('.title a')
            rank_element = await post.query_selector('.rank')

            # Extract the data we want from the elements.
            title = await title_element.inner_text() if title_element else None
            rank = await rank_element.inner_text() if rank_element else None
            href = await title_element.get_attribute('href') if title_element else None

            data.append({'title': title, 'rank': rank, 'href': href})

        # Push the extracted data to the default dataset. In local configuration,
        # the data will be stored as JSON files in ./storage/datasets/default.
        await context.push_data(data)

        # Find a link to the next page and enqueue it if it exists.
        await context.enqueue_links(selector='.morelink')

    # Define a hook that will be called each time before navigating to a new URL.
    # The hook receives a context parameter, providing access to the request and
    # browser page among other things. In this example, we log the URL being
    # navigated to.
    @crawler.pre_navigation_hook
    async def log_navigation_url(context: PlaywrightPreNavCrawlingContext) -> None:
        context.log.info(f'Navigating to {context.request.url} ...')

    # Run the crawler with the initial list of URLs.
    await crawler.run(['https://news.ycombinator.com/'])


if __name__ == '__main__':
    asyncio.run(main())



import asyncio
import re

from crawlee.crawlers import ParselCrawler, ParselCrawlingContext
from crawlee.http_clients import ImpitHttpClient
from crawlee.request_loaders import SitemapRequestLoader


async def main() -> None:
    # Create an HTTP client for fetching the sitemap.
    http_client = ImpitHttpClient()

    # Create a sitemap request loader with filtering rules.
    sitemap_loader = SitemapRequestLoader(
        sitemap_urls=['https://crawlee.dev/sitemap.xml'],
        http_client=http_client,
        include=[re.compile(r'.*docs.*')],  # Only include URLs containing 'docs'.
        max_buffer_size=500,  # Keep up to 500 URLs in memory before processing.
    )

    # highlight-start
    # Convert the sitemap loader into a request manager linked
    # to the default request queue.
    request_manager = await sitemap_loader.to_tandem()
    # highlight-end

    # Create a crawler and pass the request manager to it.
    crawler = ParselCrawler(
        request_manager=request_manager,
        max_requests_per_crawl=10,  # Limit the max requests per crawl.
    )

    @crawler.router.default_handler
    async def handler(context: ParselCrawlingContext) -> None:
        context.log.info(f'Processing {context.request.url}')

        # New links will be enqueued directly to the queue.
        await context.enqueue_links()

        # Extract data using Parsel's XPath and CSS selectors.
        data = {
            'url': context.request.url,
            'title': context.selector.xpath('//title/text()').get(),
        }

        # Push extracted data to the dataset.
        await context.push_data(data)

    await crawler.run()


if __name__ == '__main__':
    asyncio.run(main())



import asyncio

from crawlee.crawlers import BeautifulSoupCrawler, BeautifulSoupCrawlingContext


async def main() -> None:
    crawler = BeautifulSoupCrawler(
        # Limit the crawl to max requests. Remove or increase it for crawling all links.
        max_requests_per_crawl=10,
    )

    # Define the default request handler, which will be called for every request.
    @crawler.router.default_handler
    async def request_handler(context: BeautifulSoupCrawlingContext) -> None:
        context.log.info(f'Processing {context.request.url} ...')

        # Extract data from the page.
        data = {
            'url': context.request.url,
            'title': context.soup.title.string if context.soup.title else None,
        }

        # Enqueue all links found on the page.
        await context.enqueue_links()

        # Push the extracted data to the default dataset.
        await context.push_data(data)

    # Run the crawler with the initial list of URLs.
    await crawler.run(['https://crawlee.dev'])

    # Export the entire dataset to a CSV file.
    # Use semicolon as delimiter and always quote strings.
    await crawler.export_data(path='results.csv', delimiter=';', quoting='all')


if __name__ == '__main__':
    asyncio.run(main())



import asyncio

from crawlee.crawlers import PlaywrightCrawler

from .routes import router


async def main() -> None:
    crawler = PlaywrightCrawler(
        # Let's limit our crawls to make our tests shorter and safer.
        max_requests_per_crawl=10,
        # Provide our router instance to the crawler.
        request_handler=router,
    )

    await crawler.run(['https://warehouse-theme-metal.myshopify.com/collections'])


if __name__ == '__main__':
    asyncio.run(main())



import asyncio
from pathlib import Path

from crawlee.crawlers import PlaywrightCrawler, PlaywrightCrawlingContext

# Replace this with your actual Firefox profile name
# Find it at about:profiles in Firefox
PROFILE_NAME = 'your-profile-name-here'

# Paths to Firefox profiles in your system (example for Windows)
# Use `about:profiles` to find your profile path
PROFILE_PATH = Path(
    Path.home(), 'AppData', 'Roaming', 'Mozilla', 'Firefox', 'Profiles', PROFILE_NAME
)


async def main() -> None:
    crawler = PlaywrightCrawler(
        # Use Firefox browser type
        browser_type='firefox',
        # Disable fingerprints to use the profile as is
        fingerprint_generator=None,
        headless=False,
        # Path to your Firefox profile
        user_data_dir=PROFILE_PATH,
        browser_launch_options={
            'args': [
                # Required to avoid version conflicts
                '--allow-downgrade'
            ]
        },
    )

    @crawler.router.default_handler
    async def default_handler(context: PlaywrightCrawlingContext) -> None:
        context.log.info(f'Visiting {context.request.url}')

    await crawler.run(['https://crawlee.dev/'])


if __name__ == '__main__':
    asyncio.run(main())



from typing import Any

import numpy as np
from sklearn.linear_model import LogisticRegression


def sklearn_model_validator(v: LogisticRegression | dict[str, Any]) -> LogisticRegression:
    if isinstance(v, LogisticRegression):
        return v

    model = LogisticRegression(max_iter=1000)
    if v.get('is_fitted', False):
        model.coef_ = np.array(v['coef'])
        model.intercept_ = np.array(v['intercept'])
        model.classes_ = np.array(v['classes'])
        model.n_iter_ = np.array(v.get('n_iter', [1000]))

    return model


def sklearn_model_serializer(model: LogisticRegression) -> dict[str, Any]:
    if hasattr(model, 'coef_'):
        return {
            'coef': np.asarray(model.coef_).tolist(),
            'intercept': model.intercept_.tolist(),
            'classes': model.classes_.tolist(),
            'n_iter': model.n_iter_.tolist() if hasattr(model, 'n_iter_') else [1000],
            'is_fitted': True,
            'max_iter': model.max_iter,
            'solver': model.solver,
        }
    return {'is_fitted': False, 'max_iter': model.max_iter, 'solver': model.solver}
